#!/usr/bin/python3
"""."""
# -*- coding: utf-8 -*-

import settings

from mpd import MPDClient
from mpd.base import ConnectionError
from RPLCD.i2c import CharLCD
from time import sleep

from contextlib import contextmanager
from unidecode import unidecode


HOST = getattr(settings, "HOST", "localhost")
PORT = getattr(settings, "PORT", 6600)
PASSWORD = getattr(settings, "MPD_PASSWORD", None)
CUSTOM_CHARS = getattr(settings, "CUSTOM_CHARS", dict())

client = MPDClient()


class Song():
    """Song class."""

    def __init__(self, artist="<artist>", title="<title>", album="<album>"):
        """Construct method."""
        self.artist = artist
        self.title = title
        self.album = album

    def __eq__(self, other):
        """Equal operator."""
        return (self.artist == other.artist and
                self.title == other.title and
                self.album == other.album)

    def __ne__(self, other):
        """Not equal operator."""
        return (self.artist != other.artist or
                self.title != other.title or
                self.album != other.album)


class MPDStatus():
    """MPD Status class."""

    def __init__(self, song=Song(), state=None, volume=None):
        """Construct method."""
        self.song = song
        self.state = state
        self.volume = volume

    def __eq__(self, other):
        """Equal operator."""
        return (self.song == other.song and
                self.state == other.state and
                self.volume == other.volume)

    def __ne__(self, other):
        """Not equal operator."""
        return (self.song != other.song or
                self.state != other.state or
                self.volume != other.volume)


class BaseText():
    """Base Text class."""

    def __init__(self, text):
        """Construct method."""
        self.text = text
        self.length = len(text)
        self.rendered_text = text


RTL = True
LTR = False


class ScrollableText(BaseText):
    """Scrollable Text class."""

    def __init__(self, text, size):
        """Construct method."""
        super(ScrollableText, self).__init__(text)
        self._direction = RTL
        self._scroll = (len(text) > size)
        self._size = size

        if self._scroll:
            self._start_position = 0
            self._right_stop = len(text) - size + 1
            self._rendered_text = text[0:self._size]
        else:
            self._rendered_text = text

    def update(self):
        """Update rendered text."""
        if self._scroll:
            if self._direction is RTL and \
               self._start_position + 1 > self._right_stop:
                self._inverse_direction()
            elif self._direction is LTR and \
                 self._start_position - 1 < 0:
                self._inverse_direction()

            self._update_start_position()
            self._update_rendered_text()

        return self._scroll

    def _inverse_direction(self):
        """Inverse scrolling direction."""
        if self._direction is RTL:
            self._direction = LTR
        else:
            self._direction = RTL

    def _update_start_position(self):
        """Update start position."""
        if self._direction is RTL:
            self._start_position += 1
        else:
            self._start_position -= 1

    def _update_rendered_text(self):
        """Update rendered text."""
        if self._scroll:
            self._rendered_text = self.text[
                self._start_position:self._start_position + self._size
            ]

    def render(self):
        """Return rendered text."""
        return self._rendered_text

    def get_raw_text(self):
        """Return raw text."""
        return self._text


class Placeholder():
    """Placeholder class."""

    def __init__(self, position=(0, 0), length=1):
        """Construct method."""
        self._position = position
        self._length = length
        self._changed = True

    def update(self):
        """Update text object of Placeholder."""
        pass

    def has_changed(self):
        """Return changed flag and reset it."""
        result = self._changed
        self._changed = False
        return result

    def get_position(self):
        """Return position tuple of the Placeholder."""
        return self._position

    def get_length(self):
        """Return length of the Placeholder."""
        return self._length


class TextPlaceholder(Placeholder):
    """Text Placeholder class."""

    def __init__(self, position=(0, 0), length=1, text=None):
        """Construct method."""
        super(TextPlaceholder, self).__init__(position, length)
        self._text = ScrollableText(text, self._length)

    def update(self):
        """Update text object of Placeholder."""
        self._changed |= self._text.update()

    def get_length(self):
        """Return length of the Placeholder."""
        return self._length

    def get_text(self):
        """Return the rendered text for the Placeholder."""
        return self._text.render()

    def get_raw_text(self):
        """Return the rendered text for the Placeholder."""
        return self._text.render()

    def set_text(self, text):
        """Set text content of the Placeholder."""
        self._text = ScrollableText(text, self._length)
        self._changed = True


class IconPlaceholder(Placeholder):
    """Icon Placeholder class."""

    def __init__(self, position=(0, 0), icon=0):
        """Construct method."""
        super(IconPlaceholder, self).__init__(position)
        self._icon = icon

    def get_icon(self):
        """Return the rendered text for the Placeholder."""
        return self._icon

    def set_icon(self, icon):
        """Set icon char of the Placeholder."""
        self._icon = icon
        self._changed = True


class Display(CharLCD):
    """Display class."""

    def __init__(self, *args, **kwargs):
        """Construct method."""
        super(Display, self).__init__(*args, **kwargs)
        self._placeholders = dict()

    def update(self):
        """Update all Placeholders of the Display."""
        for placeholder in self._placeholders.values():
            placeholder.update()

    def get_width(self):
        """Return screen width."""
        return 20  # self.cols

    def get_height(self):
        """Return screen height."""
        return 4  # self.rows

    def add(self, placeholder_name, placeholder):
        """Add a Placeholder to the Display."""
        self._placeholders[placeholder_name] = placeholder
        self._draw_needed = True

    def clear_placeholders(self):
        """Clear all Placeholders from the Display."""
        self._placeholders = dict()
        self._draw_needed = True

    def clear_line(self, position, length):
        """Print emptu string to clear a line on Display."""
        self.cursor_pos = position
        self.write_string(' ' * length)

    def draw(self):
        """Draw all Placeholders on the Display."""
        for placeholder in self._placeholders.values():
            if placeholder.has_changed():
                self.cursor_pos = (placeholder.get_position())

                if isinstance(placeholder, TextPlaceholder):
                    if len(placeholder.get_text()) < placeholder.get_length():
                        self.clear_line(
                            placeholder.get_position(),
                            placeholder.get_length()
                        )
                    self.cursor_pos = (placeholder.get_position())
                    self.write_string(placeholder.get_text())
                elif isinstance(placeholder, IconPlaceholder):
                    self.write(placeholder.get_icon())


@contextmanager
def connection():
    """."""
    try:
        client.connect(HOST, PORT)

        if PASSWORD is not None:
            client.password(PASSWORD)

        yield
    finally:
        client.close()
        client.disconnect()


def define_icons(display):
    """Define custom icon chars for Display."""
    for custom_char in CUSTOM_CHARS.values():
        display.create_char(custom_char['char'], custom_char['data'])


def get_state_icon(state):
    """Return icon char for current state."""
    return CUSTOM_CHARS[state]['char'] or 0


def get_mpd_info():
    """Get MPD server infos."""
    with connection():
        try:
            status = client.status()
            current_song = client.currentsong()
            song = Song(
                artist=current_song.get("artist", ""),
                title=current_song.get("title", ""),
                album=current_song.get("album", "")
            )
            return MPDStatus(
                song=song,
                state=status["state"],
                volume=status["volume"]
            )

        except ConnectionError as e:
            print("Connection issue: {}".format(e.message))
            return None


def clean_data(data):
    """Normalize unicode data in Python to remove umlauts, accents etc."""
    return unidecode(data)


def main():
    """Main function."""
    display = Display(
        getattr(settings, "I2C_EXPANDER", "PCF8574"),
        getattr(settings, "ADDRESS", 0x27)
    )

    define_icons(display)

    artist = TextPlaceholder((0, 0), display.get_width(), "<artist>")
    title = TextPlaceholder((1, 0), display.get_width(), "<title>")
    album = TextPlaceholder((2, 0), display.get_width(), "<album>")
    state = TextPlaceholder((3, 2), 10, "<state>")
    volume = TextPlaceholder((3, 16), 4, "<vol>")

    state_icon = IconPlaceholder((3, 0))
    volume_icon = IconPlaceholder((3, 15), CUSTOM_CHARS['speaker']['char'])

    display.add('artist', artist)
    display.add('title', title)
    display.add('album', album)
    display.add('state', state)
    display.add('volume', volume)
    display.add('state_icon', state_icon)
    display.add('volume_icon', volume_icon)

    display.draw()

    last_info = None

    while True:
        mpd_info = get_mpd_info()

        if last_info is None:
            if mpd_info is not None:
                if mpd_info.song is not None:
                    artist.set_text(clean_data(mpd_info.song.artist))
                    title.set_text(clean_data(mpd_info.song.title))
                    album.set_text(clean_data(mpd_info.song.album))

                state.set_text(mpd_info.state)
                volume.set_text("{:3d}%".format(int(mpd_info.volume)))

                state_icon.set_icon(get_state_icon(mpd_info.state))
        elif last_info != mpd_info:
            if mpd_info.song is not None:
                if last_info.song != mpd_info.song:
                    artist.set_text(clean_data(mpd_info.song.artist))
                    title.set_text(clean_data(mpd_info.song.title))
                    album.set_text(clean_data(mpd_info.song.album))

            if mpd_info.state != state.get_raw_text():
                state.set_text(mpd_info.state)
                state_icon.set_icon(get_state_icon(mpd_info.state))

            if mpd_info.volume != volume.get_raw_text():
                volume.set_text("{:3d}%".format(int(mpd_info.volume)))

        last_info = mpd_info

        display.update()
        display.draw()
        sleep(getattr(settings, "SCROLLING_SPEED", 0.5))


if __name__ == '__main__':
    main()
